;;; gnus-ems.el --- functions for making Gnus work under different Emacsen
;; Copyright (C) 1995 Free Software Foundation, Inc.

;; Author: Lars Magne Ingebrigtsen <larsi@ifi.uio.no>
;; Keywords: news

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:

;;; Code:

(defvar gnus-mouse-2 [mouse-2])

(defalias 'gnus-make-overlay 'make-overlay)
(defalias 'gnus-overlay-put 'overlay-put)
(defalias 'gnus-move-overlay 'move-overlay)

(eval-and-compile 
  (autoload 'gnus-xmas-define "gnus-xmas")
  (autoload 'gnus-xmas-redefine "gnus-xmas"))

;;; Mule functions.

(defun gnus-mule-truncate-string (str width)
  (let ((w (string-width str))
	(col 0) (idx 0) (p-idx 0) chr)
    (if (<= w width)
	str
      (while (< col width)
	(setq chr (aref str idx)
	      col (+ col (char-width chr))
	      p-idx idx
	      idx (+ idx (char-bytes chr))
	      ))
      (substring str 0 (if (= col width)
			   idx
			 p-idx)))))

(defun gnus-mule-cite-add-face (number prefix face)
  ;; At line NUMBER, ignore PREFIX and add FACE to the rest of the line.
  (if face
      (let ((inhibit-point-motion-hooks t)
	    from to)
	(goto-line number)
	(if (boundp 'MULE)
	    (forward-char (chars-in-string prefix))
	  (forward-char (length prefix)))
	(skip-chars-forward " \t")
	(setq from (point))
	(end-of-line 1)
	(skip-chars-backward " \t")
	(setq to (point))
	(if (< from to)
	    (gnus-overlay-put (gnus-make-overlay from to) 'face face)))))

(defun gnus-mule-max-width-function (el max-width)
  (` (let* ((val (eval (, el)))
	    (valstr (if (numberp val)
			(int-to-string val) val)))
       (if (> (length valstr) (, max-width))
	   (truncate-string valstr (, max-width))
	 valstr))))


(eval
 '(progn
    (if (string-match "XEmacs\\|Lucid" emacs-version)
	()
      ;; Added by Per Abrahamsen <amanda@iesd.auc.dk>.
      (defvar gnus-display-type 
	(condition-case nil
	    (let ((display-resource (x-get-resource ".displayType" "DisplayType")))
	      (cond (display-resource (intern (downcase display-resource)))
		    ((x-display-color-p) 'color)
		    ((x-display-grayscale-p) 'grayscale)
		    (t 'mono)))
	  (error 'mono))
	"A symbol indicating the display Emacs is running under.
The symbol should be one of `color', `grayscale' or `mono'. If Emacs
guesses this display attribute wrongly, either set this variable in
your `~/.emacs' or set the resource `Emacs.displayType' in your
`~/.Xdefaults'. See also `gnus-background-mode'.

This is a meta-variable that will affect what default values other
variables get.  You would normally not change this variable, but
pounce directly on the real variables themselves.")

      (defvar gnus-background-mode 
	(condition-case nil
	    (let ((bg-resource (x-get-resource ".backgroundMode"
					       "BackgroundMode"))
		  (params (frame-parameters)))
	      (cond (bg-resource (intern (downcase bg-resource)))
		    ((and (cdr (assq 'background-color params))
			  (< (apply '+ (x-color-values
					(cdr (assq 'background-color params))))
			     (/ (apply '+ (x-color-values "white")) 3)))
		     'dark)
		    (t 'light)))
	  (error 'light))
	"A symbol indicating the Emacs background brightness.
The symbol should be one of `light' or `dark'.
If Emacs guesses this frame attribute wrongly, either set this variable in
your `~/.emacs' or set the resource `Emacs.backgroundMode' in your
`~/.Xdefaults'.
See also `gnus-display-type'.

This is a meta-variable that will affect what default values other
variables get.  You would normally not change this variable, but
pounce directly on the real variables themselves."))

    (cond 
     ((string-match "XEmacs\\|Lucid" emacs-version)
      (gnus-xmas-define))

     ((and (not (string-match "28.9" emacs-version)) 
	   (not (string-match "29" emacs-version)))
      ;; Remove the `intangible' prop.
      (let ((props (and (boundp 'gnus-hidden-properties) 
			gnus-hidden-properties)))
	(while (and props (not (eq (car (cdr props)) 'intangible)))
	  (setq props (cdr props)))
	(and props (setcdr props (cdr (cdr (cdr props))))))
      (or (fboundp 'buffer-substring-no-properties)
	  (defun buffer-substring-no-properties (beg end)
	    (format "%s" (buffer-substring beg end)))))
   
     ((boundp 'MULE)
      (provide 'gnusutil))
     )))

(eval-and-compile
  (cond
   ((not window-system)
    (defun gnus-dummy-func (&rest args))
    (let ((funcs '(mouse-set-point set-face-foreground
				   set-face-background x-popup-menu)))
      (while funcs
	(or (fboundp (car funcs))
	    (fset (car funcs) 'gnus-dummy-func))
	(setq funcs (cdr funcs))))))
  (or (fboundp 'file-regular-p)
      (defun file-regular-p (file)
	(and (not (file-directory-p file))
	     (not (file-symlink-p file))
	     (file-exists-p file))))
  (or (fboundp 'face-list)
      (defun face-list (&rest args)))
  )

(defun gnus-ems-redefine ()
  (cond 
   ((string-match "XEmacs\\|Lucid" emacs-version)
    (gnus-xmas-redefine))

   ((boundp 'MULE)
    ;; Mule definitions
    (or (fboundp 'truncate-string)
	(fset 'truncate-string 'gnus-mule-truncate-string))
    (defalias 'gnus-truncate-string 'truncate-string)

    (fset 'gnus-summary-make-display-table (lambda () nil))
    (fset 'gnus-cite-add-face 'gnus-mule-cite-add-face)
    (fset 'gnus-max-width-function 'gnus-mule-max-width-function)
    
    (if (boundp 'gnus-check-before-posting)
	(setq gnus-check-before-posting
	      (delq 'long-lines
		    (delq 'control-chars gnus-check-before-posting))))
    )
   ))

(provide 'gnus-ems)

;; Local Variables:
;; byte-compile-warnings: '(redefine callargs)
;; End:

;;; gnus-ems.el ends here
